"""UWSGI websocket proxy."""
from urllib.parse import urlparse, urlunparse
import logging
import time

import uwsgi
from pyramid import httpexceptions
from ws4py import WS_VERSION
from ws4py.client import WebSocketBaseClient


#: HTTP headers we need to proxy to upstream websocket server when the Connect: upgrade is performed
CAPTURE_CONNECT_HEADERS = ["sec-websocket-extensions", "sec-websocket-key", "origin"]


logger = logging.getLogger(__name__)


class ProxyClient(WebSocketBaseClient):
    """Proxy between upstream WebSocket server and downstream UWSGI."""

    @property
    def handshake_headers(self):
        """
        List of headers appropriate for the upgrade
        handshake.
        """
        headers = [
            ('Host', self.host),
            ('Connection', 'Upgrade'),
            ('Upgrade', 'WebSocket'),
            ('Sec-WebSocket-Key', self.key.decode('utf-8')),
            # Origin is proxyed from the downstream server, don't set it twice
            # ('Origin', self.url),
            ('Sec-WebSocket-Version', str(max(WS_VERSION)))
            ]

        if self.protocols:
            headers.append(('Sec-WebSocket-Protocol', ','.join(self.protocols)))

        if self.extra_headers:
            headers.extend(self.extra_headers)

        logger.info("Handshake headers: %s", headers)
        return headers

    def received_message(self, m):
        """Push upstream messages to downstream."""

        # TODO: No support for binary messages
        m = str(m)
        logger.debug("Incoming upstream WS: %s", m)
        uwsgi.websocket_send(m)
        logger.debug("Send ok")

    def handshake_ok(self):
        """
        Called when the upgrade handshake has completed
        successfully.

        Starts the client's thread.
        """
        self.run()

    def terminate(self):
        super(ProxyClient, self).terminate()

    def run(self):
        """Combine async uwsgi message loop with ws4py message loop.

        TODO: This could do some serious optimizations and behave asynchronously correct instead of just sleep().
        """

        self.sock.setblocking(False)
        try:
            while not self.terminated:
                logger.debug("Doing nothing")
                time.sleep(0.050)

                logger.debug("Asking for downstream msg")
                msg = uwsgi.websocket_recv_nb()
                if msg:
                    logger.debug("Incoming downstream WS: %s", msg)
                    self.send(msg)

                s = self.stream

                self.opened()

                logger.debug("Asking for upstream msg")
                try:
                    bytes = self.sock.recv(self.reading_buffer_size)
                    if bytes:
                        self.process(bytes)
                except BlockingIOError:
                    pass

        except Exception as e:
            logger.exception(e)
        finally:
            logger.info("Terminating WS proxy loop")
            self.terminate()


def serve_websocket(request, port):
    """Start UWSGI websocket loop and proxy."""
    env = request.environ

    # Send HTTP response 101 Switch Protocol downstream
    uwsgi.websocket_handshake(env['HTTP_SEC_WEBSOCKET_KEY'], env.get('HTTP_ORIGIN', ''))

    # Map the websocket URL to the upstream localhost:4000x Notebook instance
    parts = urlparse(request.url)
    parts = parts._replace(scheme="ws", netloc="localhost:{}".format(port))
    url = urlunparse(parts)

    # Proxy initial connection headers
    headers = [(header, value) for header, value in request.headers.items() if header.lower() in CAPTURE_CONNECT_HEADERS]

    logger.info("Connecting to upstream websockets: %s, headers: %s", url, headers)

    ws = ProxyClient(url, headers=headers)
    ws.connect()

    # TODO: Will complain loudly about already send headers - how to abort?
    return httpexceptions.HTTPOk()


import logging
import os
from pyramid.httpexceptions import HTTPFound
from pyramid.httpexceptions import HTTPInternalServerError
from pyramid.util import DottedNameResolver

from pyramid_notebook.notebookmanager import NotebookManager
from pyramid_notebook.proxy import WSGIProxyApplication
from pyramid_notebook.utils import make_dict_hash, route_to_alt_domain


logger = logging.getLogger(__name__)


def get_notebook_manager(request):
    settings = request.registry.settings
    notebook_folder = settings["pyramid_notebook.notebook_folder"]
    manager = NotebookManager(notebook_folder)
    return manager


def proxy_it(request, port):
    """Proxy HTTP request to upstream IPython Notebook Tornado server."""

    # Check if we have websocket proxy configured
    websocket_proxy = request.registry.settings.get("pyramid_notebook.websocket_proxy", "")
    if websocket_proxy.strip():
        r = DottedNameResolver()
        websocket_proxy = r.maybe_resolve(websocket_proxy)

    if "upgrade" in request.headers.get("connection", "").lower():
        if websocket_proxy:
            return websocket_proxy(request, port)
        else:
            # If we run on localhost on pserve, we should never hit here as requests go directly to IPython Notebook kernel, not us
            raise RuntimeError("Websocket proxy support is not configured.")

    proxy_app = WSGIProxyApplication(port)
    return request.get_response(proxy_app)


def security_check(request, username):
    """Check for basic misconfiguration errors."""
    assert username, "You must give an username whose IPython Notebook session we open"


def prepare_notebook_context(request, notebook_context):
    """Fill in notebook context with default values."""

    if not notebook_context:
        notebook_context = {}

    # Override notebook Jinja templates
    if "extra_template_paths" not in notebook_context:
        notebook_context["extra_template_paths"] = [os.path.join(os.path.dirname(__file__), "server", "templates")]

    # Furious invalid state follows if we let this slip through
    assert type(notebook_context["extra_template_paths"]) == list, "Got bad extra_template_paths {}".format(notebook_context["extra_template_paths"])

    # Jinja variables
    notebook_context["jinja_environment_options"] = notebook_context.get("jinja_environment_options", {})

    assert type(notebook_context["jinja_environment_options"]) == dict

    # XXX: Following passing of global variables to Jinja templates requires Jinja 2.8.0dev+ version and is not yet supported
    # http://jinja.pocoo.org/docs/dev/api/#jinja2.Environment.globals

    # notebook_context["jinja_environment_options"]["globals"] = notebook_context["jinja_environment_options"].get("globals", {})
    # globals_ = notebook_context["jinja_environment_options"]["globals"]
    #
    # assert type(globals_) == dict
    #
    # if not "home_url" in globals_:
    #     globals_["home_url"] = request.host_url
    #
    # if not "home_title" in globals_:
    #     globals_["home_title"] = "Back to site"

    # Tell notebook to correctly address WebSockets allow origin policy
    notebook_context["allow_origin"] = route_to_alt_domain(request, request.host_url)
    notebook_context["notebook_path"] = request.route_path("notebook_proxy", remainder="")

    # Record the hash of the current parameters, so we know if this user accesses the notebook in this or different context
    if "context_hash" not in notebook_context:
        notebook_context["context_hash"] = make_dict_hash(notebook_context)


    print(notebook_context)


def launch_on_demand(request, username, notebook_context):
    """See if we have notebook already running this context and if not then launch new one."""
    security_check(request, username)

    settings = request.registry.settings

    notebook_folder = settings.get("pyramid_notebook.notebook_folder", None)
    if not notebook_folder:
        raise RuntimeError("Setting missing: pyramid_notebook.notebook_folder")

    kill_timeout = settings.get("pyramid_notebook.kill_timeout", None)
    if not kill_timeout :
        raise RuntimeError("Setting missing: pyramid_notebook.kill_timeout")

    kill_timeout = int(kill_timeout)

    if not notebook_context:
        notebook_context = {}

    # Override notebook Jinja templates
    if "extra_template_paths" not in notebook_context:
        notebook_context["extra_template_paths"] = [os.path.join(os.path.dirname(__file__), "server", "templates")]

    # Furious invalid state follows if we let this slip through
    assert type(notebook_context["extra_template_paths"]) == list, "Got bad extra_template_paths {}".format(notebook_context["extra_template_paths"])
    notebook_folder = settings.get("pyramid_notebook.notebook_folder", None)
    if not notebook_folder:
        raise RuntimeError("Setting missing: pyramid_notebook.notebook_folder")

    kill_timeout = settings.get("pyramid_notebook.kill_timeout", None)
    if not kill_timeout :
        raise RuntimeError("Setting missing: pyramid_notebook.kill_timeout")

    kill_timeout = int(kill_timeout)

    prepare_notebook_context(request, notebook_context)

    # Configure websockets
    # websocket_url = settings.get("pyramid_notebook.websocket_url")
    # assert websocket_url, "pyramid_notebook.websocket_url setting missing"
    #assert websocket_url.startswith("ws:/") or websocket_url.startswith("wss:/")

    if request.registry.settings.get("pyramid_notebook.websocket_proxy", ""):
        websocket_url = route_to_alt_domain(request, request.host_url)
        websocket_url = websocket_url.replace("http://", "ws://").replace("https://", "wss://")
        notebook_context["websocket_url"] = websocket_url
    else:
        # Connect websockets directly to localhost notebook server, do not try to proxy them
        websocket_url =  "ws://localhost:{port}/notebook/"

    # Record the hash of the current parameters, so we know if this user accesses the notebook in this or different context
    if "context_hash" not in notebook_context:
        notebook_context["context_hash"] = make_dict_hash(notebook_context)

    manager = NotebookManager(notebook_folder, kill_timeout=kill_timeout)
    notebook_info, creates = manager.start_notebook_on_demand(username, notebook_context)
    return notebook_info


def notebook_proxy(request, username):
    """Renders a IPython Notebook frame wrapper.

    Starts or reattachs ot an existing Notebook session.
    """
    security_check(request, username)

    settings = request.registry.settings
    manager = get_notebook_manager(request)
    notebook_info = manager.get_context(username)

    if not notebook_info:
        raise HTTPInternalServerError("Apparently IPython Notebook daemon process is not running for {}".format(username))

    if not "http_port" in notebook_info:
        raise RuntimeError("Notebook terminated prematurely before managed to tell us its HTTP port")

    return proxy_it(request, notebook_info["http_port"])



def launch_notebook(request, username, notebook_context):
    """Renders a IPython Notebook frame wrapper.

    Starts or reattachs ot an existing Notebook session.
    """
    # The notebook manage now tries too hard to get the port allocated for the notebook user, making it slow
    # TODO: Manage a proper state e.g. using Redis
    notebook_info = launch_on_demand(request, username, notebook_context)

    # Jump to the detault notebook
    proxy_route = request.route_url("notebook_proxy", remainder="notebooks/{}".format(notebook_info["notebook_name"]))
    proxy_route = route_to_alt_domain(request, proxy_route)

    return HTTPFound(proxy_route)


def shutdown_notebook(request, username):
    """Stop any running notebook for a user."""

    manager = get_notebook_manager(request)
    if manager.is_running(username):
        manager.stop_notebook(username)



import copy
import os
import os.path

# Courtesy of http://stackoverflow.com/questions/5884066/hashing-a-python-dictionary
def make_dict_hash(o):
  """
  Makes a hash from a dictionary, list, tuple or set to any level, that contains
  only other hashable types (including any lists, tuples, sets, and
  dictionaries).
  """

  if isinstance(o, (set, tuple, list)):
    return tuple([make_dict_hash(e) for e in o])

  elif not isinstance(o, dict):
    return hash(o)

  new_o = copy.deepcopy(o)
  for k, v in new_o.items():
    new_o[k] = make_dict_hash(v)

  return hash(tuple(frozenset(sorted(new_o.items()))))


class change_directory:
    """ChangeDirectory is a context manager that allowing  you to temporary change the working directory.

    Courtesy of http://code.activestate.com/recipes/576620-changedirectory-context-manager/
    """

    def __init__(self, directory):
        self._dir = directory
        self._cwd = os.getcwd()
        self._pwd = self._cwd

    @property
    def current(self):
        return self._cwd

    @property
    def previous(self):
        return self._pwd

    @property
    def relative(self):
        c = self._cwd.split(os.path.sep)
        p = self._pwd.split(os.path.sep)
        l = min(len(c), len(p))
        i = 0
        while i < l and c[i] == p[i]:
            i += 1
        return os.path.normpath(os.path.join(*(['.'] + (['..'] * (len(c) - i)) + p[i:])))

    def __enter__(self):
        self._pwd = self._cwd
        os.chdir(self._dir)
        self._cwd = os.getcwd()
        return self

    def __exit__(self, *args):
        os.chdir(self._pwd)
        self._cwd = self._pwd



def route_to_alt_domain(request, url):
    """Route URL to a different subdomain.

    Used to rewrite URLs to point to websocket serving domain.
    """

    # Do we need to route IPython Notebook request from a different location
    alternative_domain = request.registry.settings.get("pyramid_notebook.alternative_domain", "").strip()
    if alternative_domain:
        url = url.replace(request.host_url, alternative_domain)
    return url




import logging
from pyramid.config import Configurator
from pyramid.authentication import BasicAuthAuthenticationPolicy
from pyramid.interfaces import IDebugLogger

from . import views
from . import auth


def mycheck(username, password, request):
    """Allow in any user with password "password"""""
    pwd_ok = password == "password"
    if not pwd_ok:
        return None
    return ['authenticated']


def main(global_config, **settings):

    settings["global_config"] = global_config
    config = Configurator(settings=settings)

    # Jinja 2 templates as .html files
    config.include('pyramid_jinja2')
    config.add_jinja2_renderer('.html')
    config.add_jinja2_search_path('pyramid_notebook:demo/templates', name='.html')

    config.add_route('home', '/')
    config.add_route('login', '/login')
    config.add_route('shell1', '/shell1')
    config.add_route('shell2', '/shell2')
    config.add_route('shutdown_notebook', '/notebook/shutdown')
    config.add_route('notebook_proxy', '/notebook/*remainder')

    config.scan(views)

    authn_policy = auth.AuthTktAuthenticationPolicy('seekrITT', callback=auth.groupfinder)
    config.set_authentication_policy(BasicAuthAuthenticationPolicy(mycheck))
    config.set_authorization_policy(authn_policy)

    # Make sure we can target Pyramid router debug messages in logging configuration
    pyramid_debug_logger = logging.getLogger("pyramid_debug")
    config.registry.registerUtility(pyramid_debug_logger, IDebugLogger)

    return config.make_wsgi_app()





"""Daemonized Python Notebook process with pre-allocated port, kill timeout and extra argument passing through JSON file."""
from contextlib import redirect_stdout
import logging

import json
import time
from IPython.nbformat.v4.nbjson import JSONWriter
import io
import os
import atexit
import signal
import sys
import faulthandler

import daemonocle
from daemonocle import expose_action
from daemonocle import DaemonError
import psutil
from pyramid_notebook.server import comm

try:
    import coverage
    coverage.process_startup()
except ImportError:
    # http://nedbatchelder.com/code/coverage/subprocess.html
    pass

port = None
kill_timeout = None
extra_argv = None
pid_file = None


class NotebookDaemon(daemonocle.Daemon):

    def __init__(self, **kwargs):
        super(NotebookDaemon, self).__init__(**kwargs)

    @expose_action
    def stop(self):
        """Stop the daemon.

        IPython Notebook tends to hang on exit 1) on certain Linux servers 2) sometimes.
        I am not sure why, but here is the traceback back when gdb was attached to the process::

            #0  0x00007fa7632c912d in poll () at ../sysdeps/unix/syscall-template.S:81
            #1  0x00007fa75e6d2f6a in poll (__timeout=<optimized out>, __nfds=2, __fds=0x7fffd2c60940) at /usr/include/x86_64-linux-gnu/bits/poll2.h:46
            #2  zmq_poll (items_=items_@entry=0x2576f00, nitems_=nitems_@entry=2, timeout_=timeout_@entry=2997) at bundled/zeromq/src/zmq.cpp:736
            #3  0x00007fa75d0d7c0b in __pyx_pf_3zmq_7backend_6cython_5_poll_zmq_poll (__pyx_self=<optimized out>, __pyx_v_timeout=2997, __pyx_v_sockets=0x7fa75b82c848) at zmq/backend/cython/_poll.c:1552
            #4  __pyx_pw_3zmq_7backend_6cython_5_poll_1zmq_poll (__pyx_self=<optimized out>, __pyx_args=<optimized out>, __pyx_kwds=<optimized out>) at zmq/backend/cython/_poll.c:1023
            #5  0x000000000057bf33 in PyEval_EvalFrameEx ()
            #6  0x000000000057d3d3 in PyEval_EvalCodeEx ()

        Smells like pyzmq bug. In any it would take pretty extensive debugging to find out why it doesn't always quit cleanly, so we just SIGKILL the process after certain timeout.

        Related bug, but this should be apparently fixed: https://github.com/zeromq/pyzmq/pull/618
        """
        if self.pidfile is None:
            raise DaemonError('Cannot stop daemon without PID file')

        pid = self._read_pidfile()
        if pid is None:
            # I don't think this should be a fatal error
            self._emit_warning('{prog} is not running'.format(prog=self.prog))
            return

        self._emit_message('Stopping {prog} ... '.format(prog=self.prog))

        try:
            # Try to terminate the process
            os.kill(pid, signal.SIGTERM)
        except OSError as ex:
            self._emit_failed()
            self._emit_error(str(ex))
            sys.exit(1)

        _, alive = psutil.wait_procs([psutil.Process(pid)], timeout=self.stop_timeout)
        if alive:
            # The process didn't terminate for some reason
            os.kill(pid, signal.SIGKILL)
            time.sleep(0.5)
            # Hahahaha. Do you feel alive now?

        self._emit_ok()

def create_named_notebook(fname, context):
    """Create a named notebook if one doesn't exist."""

    if os.path.exists(fname):
        return

    from IPython.nbformat import v4 as nbf

    # Courtesy of http://nbviewer.ipython.org/gist/fperez/9716279
    text = "Welcome to *pyramid_notebook!* -Mikko, https://opensourcehacker.com"
    cells = [nbf.new_markdown_cell(text)]

    greeting = context.get("greeting")
    if greeting:
        cells.append(nbf.new_markdown_cell(greeting))

    cells.append(nbf.new_code_cell(''))

    nb = nbf.new_notebook(cells=cells)

    with open(fname, 'w') as f:
        writer = JSONWriter()
        writer.write(nb, f)


def run_notebook(foreground=False):

    if not foreground:
        # Make it possible to get output what daemonized IPython is doing
        sys.stdout = io.open("notebook.stdout.log", "wt")
        sys.stderr = io.open("notebook.stderr.log", "wt")

    try:
        _run_notebook(foreground)
    except:
        import traceback
        traceback.print_exc(file=sys.stderr)
        raise


def _run_notebook(foreground=False):

    print("Starting notebook, daemon {}".format(not foreground), file=sys.stderr)

    # Set dead man's switch
    def kill_me(num, stack):
        """Oblig. Alien reference."""
        sys.exit(66)

    signal.signal(signal.SIGALRM, kill_me)
    signal.alarm(kill_timeout)

    argv = ["notebook", "--debug"] + extra_argv

    assert port

    os.environ["IPYTHONDIR"] = os.path.join(os.getcwd(), ".ipython")

    # Update context file with command line port settings
    context = comm.get_context(pid_file, daemon=True)

    if foreground:
        if not context:
            context = {}
    else:
        if not context:
            # We cannot let daemons start up with context, because it keeps running, reverses port but would do all proxy setup wrong
            sys.exit("Daemonized process needs an explicit context.json file and could not read one from {}".format(os.path.dirname(pid_file)))

        if "terminated" in context:
            print("Invalid context file {}: {}".format(os.path.dirname(pid_file), context), file=sys.stderr)
            sys.exit("Bad context - was by terminated notebook daemon")

        print("Starting with context {}".format(context), file=sys.stderr)

    hash = context["context_hash"]
    if hash < 0:
        # Python hasher can create negative integers which look funny in URL.
        # Let's sacrifice one bit of hash collision for aesthetics.
        hash = -hash

    notebook_name = "default-{}.ipynb".format(hash)

    context["http_port"] = port
    context["pid"] = os.getpid()
    context["kill_timeout"] = kill_timeout
    context["notebook_name"] = notebook_name

    comm.set_context(pid_file, context)

    create_named_notebook(notebook_name, context)

    # Grind through with print as IPython launcher would mess our loggers
    print("Launching on localhost:{}, having context {}".format(port, str(context)), file=sys.stderr)

    try:
        import IPython

        config = IPython.Config()

        config.NotebookApp.port = port
        config.NotebookApp.open_browser = foreground
        config.NotebookApp.base_url = context.get("notebook_path", "/notebook")
        config.NotebookApp.log_level = logging.DEBUG
        config.NotebookApp.allow_origin = context.get("allow_origin", "http://localhost:{}/".format(port))
        config.NotebookApp.extra_template_paths = context.get("extra_template_paths", [])

        if "websocket_url" in context:
            websocket_url = context.get("websocket_url", "http://localhost:{}/".format(port))
            config.NotebookApp.websocket_url = websocket_url


        if "startup" in context:
            # Drop in custom startup script
            startup_folder = os.path.join(os.getcwd(), ".ipython/profile_default/startup/")
            os.makedirs(startup_folder, exist_ok=True)
            startup_py = os.path.join(startup_folder, "startup.py")
            print("Dropping startup script {}".format(startup_py), file=sys.stderr)
            with open(startup_py, "wt") as f:
                f.write(context["startup"])

        ipython = IPython.start_ipython(argv=argv, config=config)

    except Exception as e:
        import traceback ; traceback.print_exc()
        sys.exit(str(e))


def clear_context(*args):
    comm.clear_context(pid_file)


if __name__ == '__main__':

    if len(sys.argv) == 1:
        sys.exit("Usage: {} start|stop|status|fg pid_file [work_folder] [notebook port] [kill timeout in seconds] *extra_args")

    action = sys.argv[1]
    pid_file = sys.argv[2]

    if action in ("start", "restart", "fg"):
        workdir = sys.argv[3]
        port = int(sys.argv[4])
        kill_timeout = int(sys.argv[5])
        extra_argv = sys.argv[6:]
    else:
        workdir = os.getcwd()

    os.makedirs(workdir, exist_ok=True)

    if action == "fg":
        # Test run on foreground
        os.chdir(workdir)
        atexit.register(clear_context)
        run_notebook(foreground=True)
    else:

        def shutdown(message, code):
            f = io.open("/tmp/notebook.shutdown.dump", "wt")
            faulthandler.dump_traceback(f)
            print("shutdown {}: {}".format(message, code), file=sys.stderr)
            sys.stderr.flush()
            clear_context()

        f = io.open("/tmp/notebook.dump", "wt")
        faulthandler.enable(f)

        daemon = NotebookDaemon(pidfile=pid_file, workdir=workdir, shutdown_callback=shutdown)
        daemon.worker = run_notebook
        daemon.do_action(action)



"""Communicate extra parameters to the notebook through a JSON file.

    When the notebook daemon is started, dump all special settings, including ports, in a named file similar to PID file.

    Note that port is delivered through command line AND context file. This is to allow starting daemon from the command line for testing without cumbersome writing of context file first.


"""
import datetime
import json
import shutil
import os
import logging


logger = logging.getLogger(__name__)


# http://stackoverflow.com/a/568285/315168
def check_pid(pid):
    """ Check For the existence of a unix pid. """
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


def get_context_file_name(pid_file):
    """When the daemon is started write out the information which port it was using."""
    root = os.path.dirname(pid_file)
    port_file = os.path.join(root, "context.json")
    return port_file


def set_context(pid_file, context_info):
    """Set context of running notebook.

    :param context_info: dict of extra context parameters, see comm.py comments
    """
    assert type(context_info) == dict

    port_file = get_context_file_name(pid_file)
    with open(port_file, "wt") as f:
        f.write(json.dumps(context_info))


def get_context(pid_file, daemon=False):
    """Get context of running notebook.

    :param daemon: Are we trying to fetch the context inside the daemon. Otherwise do the death check.

    :return: dict or None if the process is dead/not launcherd
    """
    port_file = get_context_file_name(pid_file)

    if not os.path.exists(port_file):
        return None

    with open(port_file, "rt") as f:
        json_data = f.read()
        try:
            data = json.loads(json_data)
        except ValueError as e:

            logger.error("Damaged context json data %s", json_data)
            return None

        if not daemon:
            pid = data.get("pid")
            if pid and not check_pid(int(pid)):
                # The Notebook daemon has exited uncleanly, as the PID does not point to any valid process
                return None


        return data


def clear_context(pid_file):
    """Called at exit. Delete the context file to signal there is no active notebook.

    We don't delete the whole file, but leave it around for debugging purposes. Maybe later we want to pass some information back to the web site.
    """
    return
    raise RuntimeError("Should not happen")
    fname = get_context_file_name(pid_file)
    shutil.move(fname, fname.replace("context.json", "context.old.json"))

    data = {}
    data["terminated"] = str(datetime.datetime.now(datetime.timezone.utc))
    set_context(pid_file, data)





import os

from pyramid.paster import get_app, setup_logging

from pyramid_notebook.demo import main

ini_path = os.path.join(os.path.dirname(__file__), "..", "..", "uwsgi-development.ini")
setup_logging(ini_path)

application = get_app(ini_path, 'main')




"""Helpers to set up Notebook startup imports and contexts."""
import os



#: The splash text displayed at the notebook top. TODO: WSGI application description in Pyramid docs is bad, find better one.
PYRAMID_GREETING = """
* **app** - The [WSGI application](http://docs.pylonsproject.org/docs/pyramid/en/latest/api/paster.html?highlight=wsgi%20application#pyramid.paster.get_app) object generated by bootstrapping.
* **request** - A [pyramid.request.Request](http://docs.pylonsproject.org/projects/pyramid/en/latest/api/request.html#pyramid.request.Request) object implying the current request state for your script.
* **root** - The resource [root](http://docs.pylonsproject.org/docs/pyramid/en/latest/glossary.html#term-root) of your Pyramid application. This is an object generated by the [root factory](http://docs.pylonsproject.org/docs/pyramid/en/latest/glossary.html#term-root-factory) configured in your application.
* **registry** - The application [registry](http://docs.pylonsproject.org/docs/pyramid/en/latest/glossary.html#term-application-registry) of your Pyramid application.
"""

#: startup.py bootstrap Python snippet to initialize the Pyramid application and get Pyramid specific variables into the shell
PYRAMID_BOOSTRAP = """
import os
from pyramid.paster import bootstrap
from pyramid_notebook.utils import change_directory

# Our development.ini, production.ini, etc.
config_file = '{}'
project_root_path = os.path.dirname(config_file)

# Most of Pyramid projects expect you run pserve et. al. in the project folder itself, thus config is hardwired to have project root relative paths
with change_directory(project_root_path):
    env = bootstrap(config_file)
    globals().update(env)

"""


def get_dotted_path(klass):
    return klass.__module__ + "." + klass.__name__


def get_import_statement(klass):
    mod = klass.__module__
    print(mod)
    return "from {} import {}".format(klass.__module__, klass.__name__)


def make_startup(notebook_context, config_file):
    """Populate notebook context with startup.py initialization file skeleton and greeting.

    This will set up context ``startup`` and ``greeting`` for their default values.

    :param notebook_context: Dictionary of notebook context info to be to passed to NotebookManager

    :param config_file: The current .ini file used to start up the Pyramid. This is used to pass it around to ``pyramid.paster.boostrap()`` to initialize dummy request object and such.
    """

    # Set up some default imports and variables

    nc = notebook_context

    add_greeting(nc, "\nAvailable variables and functions:")

    # http://docs.pylonsproject.org/projects/pyramid/en/1.1-branch/narr/commandline.html#writing-a-script

    if config_file is not None:
        assert type(config_file) == str, "Got bad config_file {}".format(config_file)
        config_file = os.path.abspath(config_file)
        assert os.path.exists(config_file), "Passed in bad config file: {}".format(config_file)
        add_script(nc, PYRAMID_BOOSTRAP.format(config_file))

        add_greeting(nc, PYRAMID_GREETING)

    add_script(nc, "import datetime")
    add_greeting(nc, "* **datetime** - Python [datetime module](https://docs.python.org/3.5/library/datetime.html)")

    add_script(nc, "import time")
    add_greeting(nc, "* **time** - Python [time module](https://docs.python.org/3.5/library/time.html)")

    try:
        # Commonly used with Pyramid applications
        import transaction
        add_script(nc,  "import transaction\n")
        add_greeting(nc, "* **transaction** - Zope [transaction manager](http://zodb.readthedocs.org/en/latest/transactions.html), e.g. `transaction.commit()`")

    except ImportError:
        pass


def include_sqlalchemy_models(nc, Base):
    """Include all SQLAlchemy models in the script context.

    :param nc: notebook_context dictionary
    :param Base: SQLAlchemy model Base class from where the all models inherit.
    """

    from sqlalchemy.ext.declarative.clsregistry import _ModuleMarker

    # Include all SQLAlchemy models in the local namespace
    for name, klass in Base._decl_class_registry.items():
        print(name, klass)
        if isinstance(klass, _ModuleMarker):
            continue

        add_script(nc, get_import_statement(klass))
        add_greeting(nc, "* **{}** - {}".format(klass.__name__, get_dotted_path(klass)))


def add_script(nc, line):
    """Add one-liner script or several lines (newline separated)"""

    assert type(nc) == dict

    nc["startup"] = nc.get("startup") or ""

    if not nc["startup"].endswith("\n"):
        nc["startup"] += "\n"
    nc["startup"] += line + "\n"


def add_greeting(nc, line):
    """Add one-liner script or several lines (newline separated)"""

    assert type(nc) == dict

    nc["greeting"] = nc.get("greeting") or ""

    # Markdown hard line break is two new lines
    if not nc["greeting"].endswith("\n"):
        nc["greeting"] += "\n"
    nc["greeting"] += line + "\n"







from pyramid import httpexceptions
from pyramid.interfaces import IAuthenticationPolicy
from pyramid.security import forget
from pyramid.view import view_config, forbidden_view_config
from pyramid.httpexceptions import HTTPUnauthorized, HTTPFound


from pyramid_notebook.views import launch_notebook
from pyramid_notebook.views import shutdown_notebook as _shutdown_notebook
from pyramid_notebook.views import notebook_proxy as _notebook_proxy
from pyramid_notebook import startup


SCRIPT = """
a = "foo"
b = "bar"

def fn():
    print("buu")
"""

GREETING="""
* **a** - varible a
* **b** - variable b
* **fn** - test function
"""


@view_config(route_name="home", renderer="templates/home.html")
def home(request):
    return {}


@view_config(route_name="notebook_proxy", renderer="templates/login.html")
def notebook_proxy(request):

    # Make sure we have a logged in user
    auth = request.registry.queryUtility(IAuthenticationPolicy)
    username = auth.authenticated_userid(request)

    if not username:
        # This will trigger HTTP Basic Auth dialog, as per basic_challenge handler below
        raise httpexceptions.HTTPForbidden("You need to be logged in. Hint: user / password")


    return _notebook_proxy(request, username)


@view_config(route_name="shell1")
def shell1(request):

    # Make sure we have a logged in user
    auth = request.registry.queryUtility(IAuthenticationPolicy)
    username = auth.authenticated_userid(request)

    if not username:
        # This will trigger HTTP Basic Auth dialog, as per basic_challenge handler below
        raise httpexceptions.HTTPForbidden("You need to be logged in. Hint: user / password")

    notebook_context = {"greeting": "**Executing shell1 context**\n"}
    config_file = request.registry.settings["global_config"]["__file__"]
    startup.make_startup(notebook_context, config_file)

    return launch_notebook(request, username, notebook_context=notebook_context)


@view_config(route_name="shell2")
def shell2(request):

    # Make sure we have a logged in user
    auth = request.registry.queryUtility(IAuthenticationPolicy)
    username = auth.authenticated_userid(request)

    if not username:
        # This will trigger HTTP Basic Auth dialog, as per basic_challenge handler below
        raise httpexceptions.HTTPForbidden("You need to be logged in. Hint: user / password")

    notebook_context = {"greeting": "**Executing shell2 context**\n"}
    config_file = request.registry.settings["global_config"]["__file__"]
    startup.make_startup(notebook_context, config_file)
    startup.add_script(notebook_context, SCRIPT)
    startup.add_greeting(notebook_context, GREETING)

    return launch_notebook(request, username, notebook_context=notebook_context)


@view_config(route_name="shutdown_notebook")
def shutdown_notebook(request):

    # Make sure we have a logged in user
    auth = request.registry.queryUtility(IAuthenticationPolicy)
    username = auth.authenticated_userid(request)

    if not username:
        # This will trigger HTTP Basic Auth dialog, as per basic_challenge handler below
        raise httpexceptions.HTTPForbidden("You need to be logged in. Hint: user / password")

    _shutdown_notebook(request, username)

    return HTTPFound(request.route_url("home"))


@forbidden_view_config()
def basic_challenge(request):
    response = HTTPUnauthorized()
    response.headers.update(forget(request))
    return response


# Courtesy of https://bitbucket.org/dahlia/wsgi-proxy/raw/02ab0dfa8e0078add268e91426e1cc1a52664cf5/wsgi_proxy/__init__.py

import http.client
import logging
from urllib.parse import urlparse, urlunsplit, urlunparse, unquote_plus


#: (:class:`frozenset`) The set of hop-by-hop headers.  All header names
#: all normalized to lowercase.
HOPPISH_HEADERS = frozenset([
    'keep-alive', 'proxy-authenticate',
    'proxy-authorization', 'te', 'trailers', 'transfer-encoding',
    'proxy-connection'
    # "upgrade", "connection"
])


def is_hop_by_hop(header):
    """Returns :const:`True` if the given ``header`` is hop by hop.

    :param header: the header name
    :type header: :class:`basestring`
    :returns: whether the given ``header`` is hop by hop or not
    :rtype: :class:`bool`

    """
    return header.lower() in HOPPISH_HEADERS


def reconstruct_url(environ, port):
    """Reconstruct the remote url from the given WSGI ``environ`` dictionary.

    :param environ: the WSGI environment
    :type environ: :class:`collections.MutableMapping`
    :returns: the remote url to proxy
    :rtype: :class:`basestring`

    """
    # From WSGI spec, PEP 333
    url = environ.get('PATH_INFO', '')
    if not url.startswith(('http://', 'https://')):
        url = '%s://%s%s' % (
            environ['wsgi.url_scheme'],
            environ['HTTP_HOST'],
            url
        )
    # Fix ;arg=value in url
    if '%3B' in url:
        url, arg = url.split('%3B', 1)
        url = ';'.join([url, arg.replace('%3D', '=')])
    # Stick query string back in
    try:
        query_string = environ['QUERY_STRING']
    except KeyError:
        pass
    else:
        url += '?' + query_string


    parsed = urlparse(url)
    replaced = parsed._replace(netloc="localhost:{}".format(port))
    url = urlunparse(replaced)
    environ['reconstructed_url'] = url
    return url


class WSGIProxyApplication:
    """WSGI application to handle requests that need to be proxied.
    You have to instantiate the class before using it as WSGI app::

        from wsgiref.simple_server import make_server

        app = WSGIProxyApplication()
        make_server('', 8080, app).serve_forever()

    """

    #: (:class:`types.ClassType`) The connection class of :mod:`httplib` module.
    #: It should be a subtype of :class:`httplib.HTTPConnection`.
    #: Default is :class:`httplib.HTTPConnection`.
    connection_class = http.client.HTTPConnection

    def __init__(self, port):
        # Target port where we proxy IPython Notebook
        self.port = port

    def handler(self, environ, start_response):
        """Proxy for requests to the actual http server"""
        logger = logging.getLogger(__name__ + '.WSGIProxyApplication.handler')
        url = urlparse(reconstruct_url(environ, self.port))

        # Create connection object
        try:
            connection = self.connection_class(url.netloc)
            # Build path
            path = url.geturl().replace('%s://%s' % (url.scheme, url.netloc),
                                        '')
        except Exception:
            start_response('501 Gateway Error', [('Content-Type', 'text/html')])
            logger.exception('Could not Connect')
            yield '<H1>Could not connect</H1>'
            return

        # Read in request body if it exists
        body = length = None
        try:
            length = int(environ['CONTENT_LENGTH'])
        except (KeyError, ValueError):

            # This is a situation where client HTTP POST is missing content-length.
            # This is also situation where (WebOb?) may screw up encoding and isert extranous = in the body.
            # https://github.com/ipython/ipython/issues/8416
            if environ["REQUEST_METHOD"] == "POST":
                if environ.get("CONTENT_TYPE") == 'application/x-www-form-urlencoded; charset=UTF-8':
                    body = environ['wsgi.input'].read()
                    try:
                        body = unquote_plus(body.decode("utf-8"))

                        # Fix extra = at end of JSON payload
                        if body.startswith("{") and body.endswith("}="):
                            body = body[0:len(body)-1]

                    except Exception as e:
                        logger.exception(e)
                        logger.error("Could not decode body: %s", body)

                    length = len(body)
        else:
            body = environ['wsgi.input'].read(length)

        # Build headers
        logger.debug('environ = %r', environ)
        headers = dict(
            (key, value)
            for key, value in (
                # This is a hacky way of getting the header names right
                (key[5:].lower().replace('_', '-'), value)
                for key, value in environ.items()
                # Keys that start with HTTP_ are all headers
                if key.startswith('HTTP_')
            )
            if not is_hop_by_hop(key)
        )

        # Handler headers that aren't HTTP_ in environ
        try:
            headers['content-type'] = environ['CONTENT_TYPE']
        except KeyError:
            pass

        # Add our host if one isn't defined
        if 'host' not in headers:
            headers['host'] = environ['SERVER_NAME']

        # Make the remote request
        try:

            logger.debug('%s %s %r',
                         environ['REQUEST_METHOD'], path, headers)
            connection.request(environ['REQUEST_METHOD'], path,
                               body=body, headers=headers)
        except Exception as e:
            # We need extra exception handling in the case the server fails
            # in mid connection, it's an edge case but I've seen it
            logger.exception(e)
            start_response('501 Gateway Error', [('Content-Type', 'text/html')])
            yield '<H1>Could not proxy IPython Notebook running localhost:{}</H1>'.format(self.port).encode("utf-8")
            return

        response = connection.getresponse()

        hopped_headers = response.getheaders()
        headers = [(key, value)
                   for key, value in hopped_headers
                   if not is_hop_by_hop(key)]

        start_response('{0.status} {0.reason}'.format(response), headers)
        while True:
            chunk = response.read(4096)
            if chunk:
                yield chunk
            else:
                break

    def __call__(self, environ, start_response):
        return self.handler(environ, start_response)




# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))


# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'CHANGES.rst'), encoding='utf-8') as f:
    long_description += "\n" + f.read()


setup(
    name='pyramid_notebook',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.1.3',

    description='Embed IPython Notebook shell on your Pyramid website',
    long_description=long_description,

    # The project's main homepage.
    url='https://bitbucket.org/miohtama/pyramid_notebook ',

    # Author details
    author='Mikko Ohtamaa',
    author_email='mikko@opensourcehacker.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Topic :: System :: Shells',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Framework :: Pyramid',
        'Framework :: IPython',
    ],

    # What does your project relate to?
    keywords='ipython setuptools development shell uwsgi',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    include_package_data=True,

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['port-for', "ipython[notebook]", 'pyramid', 'daemonocle'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest'],
        'test': ['codecov', 'pytest', 'pyramid_jinja2', 'paste', "pytest-splinter", "webtest", "pytest-cov"],
        'uwsgi': ['uwsgi', 'PasteDeploy', 'ws4py']
    },

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        "paste.app_factory": [
            'main = pyramid_notebook.demo:main'
        ]
    },
)



import json
import port_for
import logging
import subprocess
import os
import sys
import time

from .server import comm

logger = logging.getLogger(__name__)


class NotebookManager:
    """Manage any number of detached running Notebook instances.

    * Start instance by name, assign a port
    * Get instance by name
    * Stop instance by name
    * Pass extra parameters to IPython Nobetook
    """

    def __init__(self, notebook_folder, min_port=40000, port_range=10, kill_timeout=50, python=None):
        """
        :param notebook_folder: A folder containing a subfolder for each named IPython Notebook. The subfolder contains pid file, log file, default.ipynb and profile files.
        """
        self.min_port = min_port
        self.port_range = port_range
        self.notebook_folder = notebook_folder
        self.kill_timeout = kill_timeout

        if python:
            self.python = python
        else:
            self.python = self.discover_python()

        os.makedirs(notebook_folder, exist_ok=True)

        self.cmd = self.get_manager_cmd()

    def discover_python(self):
        """Get the Python interpreter we need to use to run our Notebook daemon."""
        python = sys.executable

        #: XXX fix this hack, uwsgi sets itself as Python
        #: Make better used Python interpreter autodiscovery
        if python.endswith("/uwsgi"):
            python = python.replace("/uwsgi", "/python")
        return python

    def get_work_folder(self, name):
        work_folder = os.path.join(self.notebook_folder, name)
        os.makedirs(work_folder, exist_ok=True)
        return work_folder

    def get_log_file(self, name):
        log_file = os.path.join(self.get_work_folder(name), "ipython.log")
        return log_file

    def get_pid(self, name):
        """Get PID file name for a named notebook."""
        pid_file = os.path.join(self.get_work_folder(name), "notebook.pid")
        return pid_file

    def get_context(self, name):
        return comm.get_context(self.get_pid(name))

    def get_manager_cmd(self):
        """Get our daemon script path."""
        cmd = os.path.abspath(os.path.join(os.path.dirname(__file__), "server", "notebook_daemon.py"))
        assert os.path.exists(cmd)
        return cmd

    def pick_port(self):
        """Pick open TCP/IP port."""
        ports = set(range(self.min_port, self.min_port + self.port_range))
        return port_for.select_random(ports)

    def get_notebook_daemon_command(self, name, action, port=0, *extra):
        """
        Assume we launch Notebook with the same Python which executed us.
        """

        return [self.python, self.cmd, action, self.get_pid(name), self.get_work_folder(name), port, self.kill_timeout] + list(extra)

    def exec_notebook_daemon_command(self, name, cmd, port=0):
        """Run a daemon script command."""
        cmd = self.get_notebook_daemon_command(name, cmd, port)

        # Make all arguments explicit strings
        cmd = [str(arg) for arg in cmd]

        logger.info("Running notebook command: %s", " ".join(cmd))

        # Add support for traceback dump on stuck
        env = os.environ.copy()
        env["PYTHONFAULTHANDLER"] = "true"

        p = subprocess.Popen(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE, env=env)
        stdout, stderr = p.communicate()

        if p.returncode != 0:
            logger.error("STDOUT: %s", stdout)
            logger.error("STDERR: %s", stderr)

            raise RuntimeError("Could not execute notebook command. Exit code: {} cmd: {}".format(p.returncode, " ".join(cmd)))

        return stdout

    def get_notebook_status(self, name):
        """Get the running named Notebook status.

        :return: None if no notebook is running, otherwise context dictionary
        """
        context = comm.get_context(self.get_pid(name))
        if not context:
            return None
        return context

    def start_notebook(self, name, context:dict, fg=False):
        """Start new IPython Notebook daemon.

        :param name: The owner of the Notebook will be *name*. He/she gets a new Notebook content folder created where all files are placed.

        :param context: Extra context information passed to the started Notebook. This must contain {context_hash:int} parameter used to identify the launch parameters for the notebook
        """
        assert context
        assert type(context) == dict
        assert "context_hash" in context
        assert type(context["context_hash"]) == int

        http_port = self.pick_port()
        assert http_port
        context = context.copy()
        context["http_port"] = http_port

        # We can't proxy websocket URLs, so let them go directly through localhost or have front end server to do proxying (nginx)
        if not "websocket_url" in context:
            context["websocket_url"] = "ws://localhost:{port}".format(port=http_port)

        if "{port}" in context["websocket_url"]:
            # Do port substituion for the websocket URL
             context["websocket_url"] =  context["websocket_url"].format(port=http_port)

        pid = self.get_pid(name)
        assert not "terminated" in context

        comm.set_context(pid, context)

        if fg:
            self.exec_notebook_daemon_command(name, "fg", port=http_port)
        else:
            self.exec_notebook_daemon_command(name, "start", port=http_port)

    def stop_notebook(self, name):
        self.exec_notebook_daemon_command(name, "stop")

    def is_running(self, name):
        status = self.get_notebook_status(name)
        if status:
            return status.get("pid") is not None
        return False

    def is_same_context(self, context_a, context_b):
        if context_a == context_b:
            return True

        context_a = context_a or {}
        context_b = context_b or {}

        return context_a.get("context_hash") == context_b.get("context_hash")


    def start_notebook_on_demand(self, name, context):
        """Start notebook if not yet running with these settings.

        Return the updated settings with a port info.

        :return: (context dict, created flag)
        """
        if self.is_running(name):

            last_context = self.get_context(name)
            logger.info("Notebook context change detected for {}".format(name))
            if not self.is_same_context(context, last_context):
                self.stop_notebook(name)
                # Make sure we don't get race condition over context.json file
                time.sleep(2.0)
            else:
                return last_context, False

        logger.info("Launching new Notebook named %s, context is %s", name, context)
        logger.info("Notebook log is %s/notebook.stderr.log", self.get_work_folder(name))
        self.start_notebook(name, context)
        time.sleep(1)
        return self.get_context(name), True





